@@ -13,9 +13,7 @@ License: ISC.
 [parsing doc]: https://github.com/RazrFalcon/roxmltree/blob/master/docs/parsing.md
 */
 
-#![cfg_attr(feature = "cargo-clippy", allow(collapsible_if))]
-
-#![doc(html_root_url = "https://docs.rs/roxmltree/0.4.1")]
+#![doc(html_root_url = "https://docs.rs/roxmltree/0.7.3")]
 
 #![forbid(unsafe_code)]
 #![warn(missing_docs)]
@@ -24,7 +22,7 @@ extern crate xmlparser;
 
 use std::borrow::Cow;
 use std::fmt;
-use std::ops::{Deref, Range};
+use std::ops::Deref;
 use std::rc::Rc;
 
 pub use xmlparser::TextPos;
@@ -40,6 +38,8 @@ pub const NS_XML_URI: &str = "http://www.w3.org/XML/1998/namespace";
 pub const NS_XMLNS_URI: &str = "http://www.w3.org/2000/xmlns/";
 
 
+type Range = std::ops::Range<usize>;
+
 /// An XML tree container.
 ///
 /// A tree consists of [`Nodes`].
@@ -49,30 +49,30 @@ pub const NS_XMLNS_URI: &str = "http://www.w3.org/2000/xmlns/";
 /// Root, Element, PI, Comment and Text.
 ///
 /// As you can see there are no XML declaration and CDATA types.
-/// The XML declaration is basically skipped, since it doesn't contains any
+/// The XML declaration is basically skipped, since it doesn't contain any
 /// valuable information (we support only UTF-8 anyway).
 /// And CDATA will be converted into a Text node as is, without
 /// any preprocessing (you can read more about it
 /// [here](https://github.com/RazrFalcon/roxmltree/blob/master/docs/parsing.md)).
 ///
-/// Also, the Text node data can be accesses from the text node itself or from
+/// Also, the Text node data can be accessed from the text node itself or from
 /// the parent element via [`Node::text()`] or [`Node::tail()`].
 ///
 /// [`Nodes`]: struct.Node.html
 /// [`Node::node_type()`]: struct.Node.html#method.node_type
 /// [`Node::text()`]: struct.Node.html#method.text
 /// [`Node::tail()`]: struct.Node.html#method.tail
-pub struct Document<'d> {
+pub struct Document<'input> {
     /// An original data.
     ///
     /// Required for `text_pos` methods.
-    text: &'d str,
-    nodes: Vec<NodeData<'d>>,
-    attrs: Vec<Attribute<'d>>,
-    namespaces: Namespaces<'d>,
+    text: &'input str,
+    nodes: Vec<NodeData<'input>>,
+    attrs: Vec<Attribute<'input>>,
+    namespaces: Namespaces<'input>,
 }
 
-impl<'d> Document<'d> {
+impl<'input> Document<'input> {
     /// Returns the root node.
     ///
     /// # Examples
@@ -82,7 +82,8 @@ impl<'d> Document<'d> {
     /// assert!(doc.root().is_root());
     /// assert!(doc.root().first_child().unwrap().has_tag_name("e"));
     /// ```
-    pub fn root<'a>(&'a self) -> Node<'a, 'd> {
+    #[inline]
+    pub fn root<'a>(&'a self) -> Node<'a, 'input> {
         Node { id: NodeId(0), d: &self.nodes[0], doc: self }
     }
 
@@ -90,7 +91,7 @@ impl<'d> Document<'d> {
     ///
     /// Unlike `root`, will return a first element node.
     ///
-    /// The root element is always exists.
+    /// The root element always exists.
     ///
     /// # Examples
     ///
@@ -98,14 +99,16 @@ impl<'d> Document<'d> {
     /// let doc = roxmltree::Document::parse("<!-- comment --><e/>").unwrap();
     /// assert!(doc.root_element().has_tag_name("e"));
     /// ```
-    pub fn root_element(&self) -> Node {
-        // `unwrap` is safe, because the `Document` is guarantee to have at least one element.
-        self.root().first_element_child().unwrap()
+    #[inline]
+    pub fn root_element<'a>(&'a self) -> Node<'a, 'input> {
+        // `expect` is safe, because the `Document` is guarantee to have at least one element.
+        self.root().first_element_child().expect("XML documents must contain a root element")
     }
 
     /// Returns an iterator over document's descendant nodes.
     ///
     /// Shorthand for `doc.root().descendants()`.
+    #[inline]
     pub fn descendants(&self) -> Descendants {
         self.root().descendants()
     }
@@ -117,20 +120,23 @@ impl<'d> Document<'d> {
     /// # Examples
     ///
     /// ```
-    /// let doc = roxmltree::Document::parse("\
+    /// use roxmltree::*;
+    ///
+    /// let doc = Document::parse("\
     /// <!-- comment -->
     /// <e/>"
     /// ).unwrap();
     ///
-    /// assert_eq!(doc.text_pos_from(10), roxmltree::TextPos::new(1, 11));
-    /// assert_eq!(doc.text_pos_from(9999), roxmltree::TextPos::new(2, 5));
+    /// assert_eq!(doc.text_pos_at(10), TextPos::new(1, 11));
+    /// assert_eq!(doc.text_pos_at(9999), TextPos::new(2, 5));
     /// ```
-    pub fn text_pos_from(&self, pos: usize) -> TextPos {
+    #[inline]
+    pub fn text_pos_at(&self, pos: usize) -> TextPos {
         xmlparser::Stream::from(self.text).gen_text_pos_from(pos)
     }
 }
 
-impl<'d> fmt::Debug for Document<'d> {
+impl<'input> fmt::Debug for Document<'input> {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         if !self.root().has_children() {
             return write!(f, "Document []");
@@ -218,9 +224,9 @@ pub enum NodeType {
 /// A processing instruction.
 #[derive(Clone, Copy, PartialEq, Debug)]
 #[allow(missing_docs)]
-pub struct PI<'d> {
-    pub target: &'d str,
-    pub value: Option<&'d str>,
+pub struct PI<'input> {
+    pub target: &'input str,
+    pub value: Option<&'input str>,
 }
 
 
@@ -231,38 +237,39 @@ pub struct PI<'d> {
 struct NodeId(usize);
 
 
-enum NodeKind<'d> {
+enum NodeKind<'input> {
     Root,
     Element {
-        tag_name: ExpandedNameOwned<'d>,
-        attributes: Range<usize>,
-        namespaces: Range<usize>,
+        tag_name: ExpandedNameOwned<'input>,
+        attributes: Range,
+        namespaces: Range,
     },
-    PI(PI<'d>),
-    Comment(&'d str),
-    Text(Cow<'d, str>),
+    PI(PI<'input>),
+    Comment(&'input str),
+    Text(Cow<'input, str>),
 }
 
 
-struct NodeData<'d> {
+struct NodeData<'input> {
     parent: Option<NodeId>,
     prev_sibling: Option<NodeId>,
     next_sibling: Option<NodeId>,
     children: Option<(NodeId, NodeId)>,
-    kind: NodeKind<'d>,
-    orig_pos: usize,
+    kind: NodeKind<'input>,
+    range: Range,
 }
 
 
 /// An attribute.
-pub struct Attribute<'d> {
-    name: ExpandedNameOwned<'d>,
-    value: Cow<'d, str>,
-    attr_pos: usize,
-    value_pos: usize,
+#[derive(Clone)]
+pub struct Attribute<'input> {
+    name: ExpandedNameOwned<'input>,
+    value: Cow<'input, str>,
+    range: Range,
+    value_range: Range,
 }
 
-impl<'d> Attribute<'d> {
+impl<'input> Attribute<'input> {
     /// Returns attribute's namespace URI.
     ///
     /// # Examples
@@ -275,6 +282,7 @@ impl<'d> Attribute<'d> {
     /// assert_eq!(doc.root_element().attributes()[0].namespace(), None);
     /// assert_eq!(doc.root_element().attributes()[1].namespace(), Some("http://www.w3.org"));
     /// ```
+    #[inline]
     pub fn namespace(&self) -> Option<&str> {
         self.name.ns.as_ref().map(Uri::as_str)
     }
@@ -291,6 +299,7 @@ impl<'d> Attribute<'d> {
     /// assert_eq!(doc.root_element().attributes()[0].name(), "a");
     /// assert_eq!(doc.root_element().attributes()[1].name(), "a");
     /// ```
+    #[inline]
     pub fn name(&self) -> &str {
         self.name.name
     }
@@ -307,46 +316,50 @@ impl<'d> Attribute<'d> {
     /// assert_eq!(doc.root_element().attributes()[0].value(), "b");
     /// assert_eq!(doc.root_element().attributes()[1].value(), "c");
     /// ```
+    #[inline]
     pub fn value(&self) -> &str {
         &self.value
     }
 
-    /// Returns attribute's name position in bytes in the original document.
+    /// Returns attribute's name range in bytes in the original document.
     ///
-    /// You can calculate a human-readable text position via [Node::text_pos_from].
+    /// You can calculate a human-readable text position via [Document::text_pos_at].
     ///
     /// ```text
     /// <e attr='value'/>
     ///    ^
     /// ```
     ///
-    /// [Node::text_pos_from]: struct.Node.html#method.text_pos_from
-    pub fn pos(&self) -> usize {
-        self.attr_pos
+    /// [Document::text_pos_at]: struct.Document.html#method.text_pos_at
+    #[inline]
+    pub fn range(&self) -> Range {
+        self.range.clone()
     }
 
-    /// Returns attribute's value position in bytes in the original document.
+    /// Returns attribute's value range in bytes in the original document.
     ///
-    /// You can calculate a human-readable text position via [Node::text_pos_from].
+    /// You can calculate a human-readable text position via [Document::text_pos_at].
     ///
     /// ```text
     /// <e attr='value'/>
     ///          ^
     /// ```
     ///
-    /// [Node::text_pos_from]: struct.Node.html#method.text_pos_from
-    pub fn value_pos(&self) -> usize {
-        self.value_pos
+    /// [Document::text_pos_at]: struct.Document.html#method.text_pos_at
+    #[inline]
+    pub fn value_range(&self) -> Range {
+        self.value_range.clone()
     }
 }
 
-impl<'d> PartialEq for Attribute<'d> {
-    fn eq(&self, other: &Attribute<'d>) -> bool {
+impl<'input> PartialEq for Attribute<'input> {
+    #[inline]
+    fn eq(&self, other: &Attribute<'input>) -> bool {
         self.name == other.name && self.value == other.value
     }
 }
 
-impl<'d> fmt::Debug for Attribute<'d> {
+impl<'input> fmt::Debug for Attribute<'input> {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         write!(f, "Attribute {{ name: {:?}, value: {:?} }}",
                self.name, self.value)
@@ -358,12 +371,12 @@ impl<'d> fmt::Debug for Attribute<'d> {
 ///
 /// Contains URI and *prefix* pair.
 #[derive(Clone, PartialEq, Debug)]
-pub struct Namespace<'d> {
-    name: Option<&'d str>,
+pub struct Namespace<'input> {
+    name: Option<&'input str>,
     uri: Uri,
 }
 
-impl<'d> Namespace<'d> {
+impl<'input> Namespace<'input> {
     /// Returns namespace name/prefix.
     ///
     /// # Examples
@@ -383,6 +396,7 @@ impl<'d> Namespace<'d> {
     ///
     /// assert_eq!(doc.root_element().namespaces()[0].name(), None);
     /// ```
+    #[inline]
     pub fn name(&self) -> Option<&str> {
         self.name
     }
@@ -398,16 +412,18 @@ impl<'d> Namespace<'d> {
     ///
     /// assert_eq!(doc.root_element().namespaces()[0].uri(), "http://www.w3.org");
     /// ```
+    #[inline]
     pub fn uri(&self) -> &str {
         self.uri.as_str()
     }
 }
 
 
-struct Namespaces<'d>(Vec<Namespace<'d>>);
+struct Namespaces<'input>(Vec<Namespace<'input>>);
 
-impl<'d> Namespaces<'d> {
-    fn push_ns(&mut self, name: Option<&'d str>, uri: String) {
+impl<'input> Namespaces<'input> {
+    #[inline]
+    fn push_ns(&mut self, name: Option<&'input str>, uri: String) {
         debug_assert_ne!(name, Some(""));
 
         self.0.push(Namespace {
@@ -416,26 +432,21 @@ impl<'d> Namespaces<'d> {
         });
     }
 
+    #[inline]
     fn xml_uri(&self) -> Uri {
         self[0].uri.clone()
     }
 
-    fn get_by_prefix(&self, range: Range<usize>, prefix: &str) -> Option<Uri> {
-        let prefix = if prefix.is_empty() { None } else { Some(prefix) };
-
-        self[range].iter()
-                   .find(|ns| ns.name == prefix)
-                   .map(|ns| ns.uri.clone())
-    }
-
+    #[inline]
     fn exists(&self, start: usize, prefix: Option<&str>) -> bool {
         self[start..].iter().any(|ns| ns.name == prefix)
     }
 }
 
-impl<'d> Deref for Namespaces<'d> {
-    type Target = Vec<Namespace<'d>>;
+impl<'input> Deref for Namespaces<'input> {
+    type Target = Vec<Namespace<'input>>;
 
+    #[inline]
     fn deref(&self) -> &Self::Target {
         &self.0
     }
@@ -445,22 +456,26 @@ impl<'d> Deref for Namespaces<'d> {
 struct Uri(Rc<String>);
 
 impl Uri {
+    #[inline]
     fn new(text: String) -> Self {
         Uri(Rc::new(text))
     }
 
+    #[inline]
     fn as_str(&self) -> &str {
         self.0.as_str()
     }
 }
 
 impl Clone for Uri {
+    #[inline]
     fn clone(&self) -> Self {
         Uri(Rc::clone(&self.0))
     }
 }
 
 impl PartialEq for Uri {
+    #[inline]
     fn eq(&self, other: &Uri) -> bool {
         self.0.as_str() == other.0.as_str()
     }
@@ -474,18 +489,23 @@ impl fmt::Debug for Uri {
 
 
 #[derive(Clone, PartialEq)]
-struct ExpandedNameOwned<'d> {
+struct ExpandedNameOwned<'input> {
     ns: Option<Uri>,
-    name: &'d str,
+    prefix: &'input str, // Used only for closing tags matching during parsing.
+    name: &'input str,
 }
 
-impl<'d> ExpandedNameOwned<'d> {
+impl<'input> ExpandedNameOwned<'input> {
+    #[inline]
     fn as_ref(&self) -> ExpandedName {
-        ExpandedName { uri: self.ns.as_ref().map(Uri::as_str), name: self.name }
+        ExpandedName {
+            uri: self.ns.as_ref().map(Uri::as_str),
+            name: self.name,
+        }
     }
 }
 
-impl<'d> fmt::Debug for ExpandedNameOwned<'d> {
+impl<'input> fmt::Debug for ExpandedNameOwned<'input> {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         match self.ns {
             Some(ref ns) => write!(f, "{{{}}}{}", ns.as_str(), self.name),
@@ -499,12 +519,12 @@ impl<'d> fmt::Debug for ExpandedNameOwned<'d> {
 ///
 /// Contains an namespace URI and name pair.
 #[derive(Clone, Copy, PartialEq)]
-pub struct ExpandedName<'d> {
-    uri: Option<&'d str>,
-    name: &'d str,
+pub struct ExpandedName<'input> {
+    uri: Option<&'input str>,
+    name: &'input str,
 }
 
-impl<'d> ExpandedName<'d> {
+impl<'input> ExpandedName<'input> {
     /// Returns a namespace URI.
     ///
     /// # Examples
@@ -514,11 +534,12 @@ impl<'d> ExpandedName<'d> {
     ///
     /// assert_eq!(doc.root_element().tag_name().namespace(), Some("http://www.w3.org"));
     /// ```
-    pub fn namespace(&self) -> Option<&'d str> {
+    #[inline]
+    pub fn namespace(&self) -> Option<&'input str> {
         self.uri
     }
 
-    /// Returns a name.
+    /// Returns a local name.
     ///
     /// # Examples
     ///
@@ -527,12 +548,13 @@ impl<'d> ExpandedName<'d> {
     ///
     /// assert_eq!(doc.root_element().tag_name().name(), "e");
     /// ```
-    pub fn name(&self) -> &'d str {
+    #[inline]
+    pub fn name(&self) -> &'input str {
         self.name
     }
 }
 
-impl<'d> fmt::Debug for ExpandedName<'d> {
+impl<'input> fmt::Debug for ExpandedName<'input> {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         match self.namespace() {
             Some(ns) => write!(f, "{{{}}}{}", ns, self.name),
@@ -541,8 +563,9 @@ impl<'d> fmt::Debug for ExpandedName<'d> {
     }
 }
 
-impl<'d> From<&'d str> for ExpandedName<'d> {
-    fn from(v: &'d str) -> Self {
+impl<'input> From<&'input str> for ExpandedName<'input> {
+    #[inline]
+    fn from(v: &'input str) -> Self {
         ExpandedName {
             uri: None,
             name: v,
@@ -550,8 +573,9 @@ impl<'d> From<&'d str> for ExpandedName<'d> {
     }
 }
 
-impl<'d> From<(&'d str, &'d str)> for ExpandedName<'d> {
-    fn from(v: (&'d str, &'d str)) -> Self {
+impl<'input> From<(&'input str, &'input str)> for ExpandedName<'input> {
+    #[inline]
+    fn from(v: (&'input str, &'input str)) -> Self {
         ExpandedName {
             uri: Some(v.0),
             name: v.1,
@@ -562,19 +586,20 @@ impl<'d> From<(&'d str, &'d str)> for ExpandedName<'d> {
 
 /// A node.
 #[derive(Clone, Copy)]
-pub struct Node<'a, 'd: 'a> {
+pub struct Node<'a, 'input: 'a> {
     /// Node ID.
     id: NodeId,
 
     /// Tree containing the node.
-    doc: &'a Document<'d>,
+    doc: &'a Document<'input>,
 
-    d: &'a NodeData<'d>,
+    d: &'a NodeData<'input>,
 }
 
-impl<'a, 'd> Eq for Node<'a, 'd> {}
+impl<'a, 'input> Eq for Node<'a, 'input> {}
 
-impl<'a, 'd> PartialEq for Node<'a, 'd> {
+impl<'a, 'input> PartialEq for Node<'a, 'input> {
+    #[inline]
     fn eq(&self, other: &Self) -> bool {
            self.id == other.id
         && self.doc as *const _ == other.doc as *const _
@@ -582,8 +607,9 @@ impl<'a, 'd> PartialEq for Node<'a, 'd> {
     }
 }
 
-impl<'a, 'd: 'a> Node<'a, 'd> {
+impl<'a, 'input: 'a> Node<'a, 'input> {
     /// Returns node's type.
+    #[inline]
     pub fn node_type(&self) -> NodeType {
         match self.d.kind {
             NodeKind::Root => NodeType::Root,
@@ -595,32 +621,38 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     }
 
     /// Checks that node is a root node.
+    #[inline]
     pub fn is_root(&self) -> bool {
         self.node_type() == NodeType::Root
     }
 
     /// Checks that node is an element node.
+    #[inline]
     pub fn is_element(&self) -> bool {
         self.node_type() == NodeType::Element
     }
 
     /// Checks that node is a processing instruction node.
+    #[inline]
     pub fn is_pi(&self) -> bool {
         self.node_type() == NodeType::PI
     }
 
     /// Checks that node is a comment node.
+    #[inline]
     pub fn is_comment(&self) -> bool {
         self.node_type() == NodeType::Comment
     }
 
     /// Checks that node is a text node.
+    #[inline]
     pub fn is_text(&self) -> bool {
         self.node_type() == NodeType::Text
     }
 
     /// Returns node's document.
-    pub fn document(&self) -> &'a Document<'d> {
+    #[inline]
+    pub fn document(&self) -> &'a Document<'input> {
         self.doc
     }
 
@@ -636,6 +668,7 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     /// assert_eq!(doc.root_element().tag_name().namespace(), Some("http://www.w3.org"));
     /// assert_eq!(doc.root_element().tag_name().name(), "e");
     /// ```
+    #[inline]
     pub fn tag_name(&self) -> ExpandedName<'a> {
         match self.d.kind {
             NodeKind::Element { ref tag_name, .. } => tag_name.as_ref(),
@@ -656,8 +689,9 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     /// assert!(!doc.root_element().has_tag_name("b"));
     /// assert!(!doc.root_element().has_tag_name(("http://www.w4.org", "e")));
     /// ```
-    pub fn has_tag_name<N>(&self, name: N) -> bool
-        where N: Into<ExpandedName<'a>>
+    pub fn has_tag_name<'n, N>(&self, name: N) -> bool
+    where
+        N: Into<ExpandedName<'n>>,
     {
         let name = name.into();
 
@@ -691,47 +725,6 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
         self.namespaces().iter().find(|ns| ns.name.is_none()).map(|v| v.uri.as_str())
     }
 
-    /// Returns element's namespace prefix.
-    ///
-    /// Returns `None`:
-    /// - if the current node is not an element
-    /// - if the current element has a default namespace
-    /// - if the current element has no namespace
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse("<n:e xmlns:n='http://www.w3.org'/>").unwrap();
-    ///
-    /// assert_eq!(doc.root_element().resolve_tag_name_prefix(), Some("n"));
-    /// ```
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse("<e xmlns:n='http://www.w3.org'/>").unwrap();
-    ///
-    /// assert_eq!(doc.root_element().resolve_tag_name_prefix(), None);
-    /// ```
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse("<e xmlns='http://www.w3.org'/>").unwrap();
-    ///
-    /// assert_eq!(doc.root_element().resolve_tag_name_prefix(), None);
-    /// ```
-    pub fn resolve_tag_name_prefix(&self) -> Option<&'a str> {
-        if !self.is_element() {
-            return None;
-        }
-
-        let tag_ns = self.tag_name();
-
-        if self.default_namespace() == tag_ns.namespace() {
-            return None;
-        }
-
-        self.namespaces().iter().find(|ns| Some(ns.uri()) == tag_ns.namespace())
-            .map(|v| v.name).unwrap_or(None)
-    }
-
     /// Returns a prefix for a given namespace URI.
     ///
     /// # Examples
@@ -792,13 +785,27 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     /// assert_eq!(doc.root_element().attribute("a"), Some("b"));
     /// assert_eq!(doc.root_element().attribute(("http://www.w3.org", "a")), Some("c"));
     /// ```
-    pub fn attribute<N>(&self, name: N) -> Option<&'a str>
-        where N: Into<ExpandedName<'a>>
+    pub fn attribute<'n, N>(&self, name: N) -> Option<&'a str>
+    where
+        N: Into<ExpandedName<'n>>,
     {
         let name = name.into();
         self.attributes().iter().find(|a| a.name.as_ref() == name).map(|a| a.value.as_ref())
     }
 
+    /// Returns element's attribute object.
+    ///
+    /// The same as [`attribute()`], but returns the `Attribute` itself instead of a value string.
+    ///
+    /// [`attribute()`]: struct.Node.html#method.attribute
+    pub fn attribute_node<'n, N>(&self, name: N) -> Option<&'a Attribute<'input>>
+    where
+        N: Into<ExpandedName<'n>>,
+    {
+        let name = name.into();
+        self.attributes().iter().find(|a| a.name.as_ref() == name)
+    }
+
     /// Checks that element has a specified attribute.
     ///
     /// # Examples
@@ -814,8 +821,9 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     /// assert!(!doc.root_element().has_attribute("b"));
     /// assert!(!doc.root_element().has_attribute(("http://www.w4.org", "a")));
     /// ```
-    pub fn has_attribute<N>(&self, name: N) -> bool
-        where N: Into<ExpandedName<'a>>
+    pub fn has_attribute<'n, N>(&self, name: N) -> bool
+    where
+        N: Into<ExpandedName<'n>>,
     {
         let name = name.into();
         self.attributes().iter().any(|a| a.name.as_ref() == name)
@@ -832,61 +840,14 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     ///
     /// assert_eq!(doc.root_element().attributes().len(), 2);
     /// ```
-    pub fn attributes(&self) -> &'a [Attribute<'d>] {
+    #[inline]
+    pub fn attributes(&self) -> &'a [Attribute<'input>] {
         match self.d.kind {
             NodeKind::Element { ref attributes, .. } => &self.doc.attrs[attributes.clone()],
             _ => &[],
         }
     }
 
-    /// Calculates attribute's position in the original document.
-    ///
-    /// **Note:** this operation is expensive.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse(
-    ///     "<e xmlns:n='http://www.w3.org' a='b' n:a='c'/>"
-    /// ).unwrap();
-    ///
-    /// assert_eq!(doc.root_element().attribute_pos("a"),
-    ///            Some(roxmltree::TextPos::new(1, 32)));
-    /// assert_eq!(doc.root_element().attribute_pos(("http://www.w3.org", "a")),
-    ///            Some(roxmltree::TextPos::new(1, 38)));
-    /// ```
-    pub fn attribute_pos<N>(&self, name: N) -> Option<TextPos>
-        where N: Into<ExpandedName<'a>>
-    {
-        let name = name.into();
-        self.attributes().iter().find(|a| a.name.as_ref() == name)
-            .map(|a| self.document().text_pos_from(a.attr_pos))
-    }
-
-    /// Calculates attribute's value position in the original document.
-    ///
-    /// **Note:** this operation is expensive.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse(
-    ///     "<e xmlns:n='http://www.w3.org' a='b' n:a='c'/>"
-    /// ).unwrap();
-    ///
-    /// assert_eq!(doc.root_element().attribute_value_pos("a"),
-    ///            Some(roxmltree::TextPos::new(1, 35)));
-    /// assert_eq!(doc.root_element().attribute_value_pos(("http://www.w3.org", "a")),
-    ///            Some(roxmltree::TextPos::new(1, 43)));
-    /// ```
-    pub fn attribute_value_pos<N>(&self, name: N) -> Option<TextPos>
-        where N: Into<ExpandedName<'a>>
-    {
-        let name = name.into();
-        self.attributes().iter().find(|a| a.name.as_ref() == name)
-            .map(|a| self.document().text_pos_from(a.value_pos))
-    }
-
     /// Returns element's namespaces.
     ///
     /// # Examples
@@ -898,7 +859,8 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     ///
     /// assert_eq!(doc.root_element().namespaces().len(), 1);
     /// ```
-    pub fn namespaces(&self) -> &'a [Namespace<'d>] {
+    #[inline]
+    pub fn namespaces(&self) -> &'a [Namespace<'input>] {
         match self.d.kind {
             NodeKind::Element { ref namespaces, .. } => {
                 &self.doc.namespaces[namespaces.clone()]
@@ -933,6 +895,7 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     ///
     /// assert_eq!(doc.root().first_child().unwrap().text(), Some(" comment "));
     /// ```
+    #[inline]
     pub fn text(&self) -> Option<&'a str> {
         match self.d.kind {
             NodeKind::Element { .. } => {
@@ -968,6 +931,7 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     /// let p = doc.descendants().find(|n| n.has_tag_name("p")).unwrap();
     /// assert_eq!(p.tail(), Some("\n    text2\n"));
     /// ```
+    #[inline]
     pub fn tail(&self) -> Option<&'a str> {
         if !self.is_element() {
             return None;
@@ -985,38 +949,54 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     }
 
     /// Returns node as Processing Instruction.
-    pub fn pi(&self) -> Option<PI<'d>> {
+    #[inline]
+    pub fn pi(&self) -> Option<PI<'input>> {
         match self.d.kind {
             NodeKind::PI(pi) => Some(pi),
             _ => None,
         }
     }
 
-    fn gen_node(&self, id: NodeId) -> Node<'a, 'd> {
+    #[inline]
+    fn gen_node(&self, id: NodeId) -> Node<'a, 'input> {
         Node { id, d: &self.doc.nodes[id.0], doc: self.doc }
     }
 
     /// Returns the parent of this node.
+    #[inline]
     pub fn parent(&self) -> Option<Self> {
         self.d.parent.map(|id| self.gen_node(id))
     }
 
     /// Returns the parent element of this node.
     pub fn parent_element(&self) -> Option<Self> {
-        self.ancestors().filter(|n| n.is_element()).nth(0)
+        self.ancestors().skip(1).filter(|n| n.is_element()).nth(0)
     }
 
     /// Returns the previous sibling of this node.
+    #[inline]
     pub fn prev_sibling(&self) -> Option<Self> {
         self.d.prev_sibling.map(|id| self.gen_node(id))
     }
 
+    /// Returns the previous sibling element of this node.
+    pub fn prev_sibling_element(&self) -> Option<Self> {
+        self.prev_siblings().filter(|n| n.is_element()).nth(0)
+    }
+
     /// Returns the next sibling of this node.
+    #[inline]
     pub fn next_sibling(&self) -> Option<Self> {
         self.d.next_sibling.map(|id| self.gen_node(id))
     }
 
+    /// Returns the next sibling element of this node.
+    pub fn next_sibling_element(&self) -> Option<Self> {
+        self.next_siblings().filter(|n| n.is_element()).nth(0)
+    }
+
     /// Returns the first child of this node.
+    #[inline]
     pub fn first_child(&self) -> Option<Self> {
         self.d.children.map(|(id, _)| self.gen_node(id))
     }
@@ -1027,6 +1007,7 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     }
 
     /// Returns the last child of this node.
+    #[inline]
     pub fn last_child(&self) -> Option<Self> {
         self.d.children.map(|(_, id)| self.gen_node(id))
     }
@@ -1037,80 +1018,73 @@ impl<'a, 'd: 'a> Node<'a, 'd> {
     }
 
     /// Returns true if this node has siblings.
+    #[inline]
     pub fn has_siblings(&self) -> bool {
         self.d.prev_sibling.is_some() || self.d.next_sibling.is_some()
     }
 
     /// Returns true if this node has children.
+    #[inline]
     pub fn has_children(&self) -> bool {
         self.d.children.is_some()
     }
 
-    /// Returns an iterator over ancestor nodes.
-    pub fn ancestors(&self) -> Ancestors<'a, 'd> {
-        Ancestors(self.parent())
+    /// Returns an iterator over ancestor nodes starting at this node.
+    #[inline]
+    pub fn ancestors(&self) -> Ancestors<'a, 'input> {
+        Ancestors(Some(*self))
     }
 
     /// Returns an iterator over previous sibling nodes.
-    pub fn prev_siblings(&self) -> PrevSiblings<'a, 'd> {
+    #[inline]
+    pub fn prev_siblings(&self) -> PrevSiblings<'a, 'input> {
         PrevSiblings(self.prev_sibling())
     }
 
     /// Returns an iterator over next sibling nodes.
-    pub fn next_siblings(&self) -> NextSiblings<'a, 'd> {
+    #[inline]
+    pub fn next_siblings(&self) -> NextSiblings<'a, 'input> {
         NextSiblings(self.next_sibling())
     }
 
     /// Returns an iterator over first children nodes.
-    pub fn first_children(&self) -> FirstChildren<'a, 'd> {
+    #[inline]
+    pub fn first_children(&self) -> FirstChildren<'a, 'input> {
         FirstChildren(self.first_child())
     }
 
     /// Returns an iterator over last children nodes.
-    pub fn last_children(&self) -> LastChildren<'a, 'd> {
+    #[inline]
+    pub fn last_children(&self) -> LastChildren<'a, 'input> {
         LastChildren(self.last_child())
     }
 
     /// Returns an iterator over children nodes.
-    pub fn children(&self) -> Children<'a, 'd> {
+    #[inline]
+    pub fn children(&self) -> Children<'a, 'input> {
         Children { front: self.first_child(), back: self.last_child() }
     }
 
     /// Returns an iterator which traverses the subtree starting at this node.
-    pub fn traverse(&self) -> Traverse<'a, 'd> {
+    #[inline]
+    pub fn traverse(&self) -> Traverse<'a, 'input> {
         Traverse { root: *self, edge: None }
     }
 
     /// Returns an iterator over this node and its descendants.
-    pub fn descendants(&self) -> Descendants<'a, 'd> {
+    #[inline]
+    pub fn descendants(&self) -> Descendants<'a, 'input> {
         Descendants(self.traverse())
     }
 
-    /// Returns node's position in bytes in the original document.
-    pub fn pos(&self) -> usize {
-        self.d.orig_pos
-    }
-
-    /// Calculates node's position in the original document.
-    ///
-    /// **Note:** this operation is expensive.
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// let doc = roxmltree::Document::parse("\
-    /// <!-- comment -->
-    /// <e/>"
-    /// ).unwrap();
-    ///
-    /// assert_eq!(doc.root_element().node_pos(), roxmltree::TextPos::new(2, 1));
-    /// ```
-    pub fn node_pos(&self) -> TextPos {
-        self.document().text_pos_from(self.d.orig_pos)
+    /// Returns node's range in bytes in the original document.
+    #[inline]
+    pub fn range(&self) -> Range {
+        self.d.range.clone()
     }
 }
 
-impl<'a, 'd: 'a> fmt::Debug for Node<'a, 'd> {
+impl<'a, 'input: 'a> fmt::Debug for Node<'a, 'input> {
     fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
         match self.d.kind {
             NodeKind::Root => write!(f, "Root"),
@@ -1132,10 +1106,12 @@ macro_rules! axis_iterators {
         $(
             #[$m]
             #[derive(Clone)]
-            pub struct $i<'a, 'd: 'a>(Option<Node<'a, 'd>>);
+            pub struct $i<'a, 'input: 'a>(Option<Node<'a, 'input>>);
 
-            impl<'a, 'd: 'a> Iterator for $i<'a, 'd> {
-                type Item = Node<'a, 'd>;
+            impl<'a, 'input: 'a> Iterator for $i<'a, 'input> {
+                type Item = Node<'a, 'input>;
+
+                #[inline]
                 fn next(&mut self) -> Option<Self::Item> {
                     let node = self.0.take();
                     self.0 = node.as_ref().and_then($f);
@@ -1166,14 +1142,15 @@ axis_iterators! {
 
 /// Iterator over children.
 #[derive(Clone)]
-pub struct Children<'a, 'd: 'a> {
-    front: Option<Node<'a, 'd>>,
-    back: Option<Node<'a, 'd>>,
+pub struct Children<'a, 'input: 'a> {
+    front: Option<Node<'a, 'input>>,
+    back: Option<Node<'a, 'input>>,
 }
 
-impl<'a, 'd: 'a> Iterator for Children<'a, 'd> {
-    type Item = Node<'a, 'd>;
+impl<'a, 'input: 'a> Iterator for Children<'a, 'input> {
+    type Item = Node<'a, 'input>;
 
+    #[inline]
     fn next(&mut self) -> Option<Self::Item> {
         if self.front == self.back {
             let node = self.front.take();
@@ -1187,7 +1164,8 @@ impl<'a, 'd: 'a> Iterator for Children<'a, 'd> {
     }
 }
 
-impl<'a, 'd: 'a> DoubleEndedIterator for Children<'a, 'd> {
+impl<'a, 'input: 'a> DoubleEndedIterator for Children<'a, 'input> {
+    #[inline]
     fn next_back(&mut self) -> Option<Self::Item> {
         if self.back == self.front {
             let node = self.back.take();
@@ -1204,24 +1182,25 @@ impl<'a, 'd: 'a> DoubleEndedIterator for Children<'a, 'd> {
 
 /// Open or close edge of a node.
 #[derive(Clone, Copy, PartialEq, Debug)]
-pub enum Edge<'a, 'd: 'a> {
+pub enum Edge<'a, 'input: 'a> {
     /// Open.
-    Open(Node<'a, 'd>),
+    Open(Node<'a, 'input>),
     /// Close.
-    Close(Node<'a, 'd>),
+    Close(Node<'a, 'input>),
 }
 
 
 /// Iterator which traverses a subtree.
 #[derive(Clone)]
-pub struct Traverse<'a, 'd: 'a> {
-    root: Node<'a, 'd>,
-    edge: Option<Edge<'a, 'd>>,
+pub struct Traverse<'a, 'input: 'a> {
+    root: Node<'a, 'input>,
+    edge: Option<Edge<'a, 'input>>,
 }
 
-impl<'a, 'd: 'a> Iterator for Traverse<'a, 'd> {
-    type Item = Edge<'a, 'd>;
+impl<'a, 'input: 'a> Iterator for Traverse<'a, 'input> {
+    type Item = Edge<'a, 'input>;
 
+    #[inline]
     fn next(&mut self) -> Option<Self::Item> {
         match self.edge {
             Some(Edge::Open(node)) => {
@@ -1251,11 +1230,12 @@ impl<'a, 'd: 'a> Iterator for Traverse<'a, 'd> {
 
 /// Iterator over a node and its descendants.
 #[derive(Clone)]
-pub struct Descendants<'a, 'd: 'a>(Traverse<'a, 'd>);
+pub struct Descendants<'a, 'input: 'a>(Traverse<'a, 'input>);
 
-impl<'a, 'd: 'a> Iterator for Descendants<'a, 'd> {
-    type Item = Node<'a, 'd>;
+impl<'a, 'input: 'a> Iterator for Descendants<'a, 'input> {
+    type Item = Node<'a, 'input>;
 
+    #[inline]
     fn next(&mut self) -> Option<Self::Item> {
         for edge in &mut self.0 {
             if let Edge::Open(node) = edge {
