@@ -2,7 +2,6 @@ use std::borrow::{Cow, Borrow};
 use std::error;
 use std::fmt;
 use std::mem;
-use std::ops::Range;
 use std::str;
 
 use xmlparser::{
@@ -20,11 +19,12 @@ use {
     Document,
     ExpandedNameOwned,
     Namespaces,
-    Node,
     NodeData,
     NodeId,
     NodeKind,
     PI,
+    Range,
+    Uri,
 };
 
 const ENTITY_DEPTH: u8 = 10;
@@ -48,6 +48,13 @@ pub enum Error {
     /// A namespace was already defined on this element.
     DuplicatedNamespace(String, TextPos),
 
+    /// An unknown namespace.
+    ///
+    /// Indicates that an element or an attribute has an unknown qualified name prefix.
+    ///
+    /// The first value is a prefix.
+    UnknownNamespace(String, TextPos),
+
     /// Incorrect tree structure.
     #[allow(missing_docs)]
     UnexpectedCloseTag { expected: String, actual: String, pos: TextPos },
@@ -85,6 +92,7 @@ pub enum Error {
 
 impl Error {
     /// Returns the error position.
+    #[inline]
     pub fn pos(&self) -> TextPos {
         match *self {
             Error::InvalidXmlPrefixUri(pos) => pos,
@@ -92,6 +100,7 @@ impl Error {
             Error::UnexpectedXmlnsUri(pos) => pos,
             Error::InvalidElementNamePrefix(pos) => pos,
             Error::DuplicatedNamespace(ref _name, pos) => pos,
+            Error::UnknownNamespace(ref _name, pos) => pos,
             Error::UnexpectedCloseTag { pos, .. } => pos,
             Error::UnexpectedEntityCloseTag(pos) => pos,
             Error::UnknownEntityReference(ref _name, pos) => pos,
@@ -104,6 +113,7 @@ impl Error {
 }
 
 impl From<xmlparser::Error> for Error {
+    #[inline]
     fn from(e: xmlparser::Error) -> Self {
         Error::ParserError(e)
     }
@@ -127,6 +137,9 @@ impl fmt::Display for Error {
             Error::DuplicatedNamespace(ref name, pos) => {
                 write!(f, "namespace '{}' at {} is already defined", name, pos)
             }
+            Error::UnknownNamespace(ref name, pos) => {
+                write!(f, "an unknown namespace prefix '{}' at {}", name, pos)
+            }
             Error::UnexpectedCloseTag { ref expected, ref actual, pos } => {
                 write!(f, "expected '{}' tag, not '{}' at {}", expected, actual, pos)
             }
@@ -153,22 +166,22 @@ impl fmt::Display for Error {
 }
 
 impl error::Error for Error {
+    #[inline]
     fn description(&self) -> &str {
         "an XML parsing error"
     }
 }
 
 
-struct AttributeData<'d> {
-    /// Contains prefix or local name for error position.
-    name: StrSpan<'d>,
-    prefix: &'d str,
-    local: &'d str,
-    value: Cow<'d, str>,
-    value_pos: usize,
+struct AttributeData<'input> {
+    prefix: StrSpan<'input>,
+    local: StrSpan<'input>,
+    value: Cow<'input, str>,
+    range: Range,
+    value_range: Range,
 }
 
-impl<'d> Document<'d> {
+impl<'input> Document<'input> {
     /// Parses the input XML string.
     ///
     /// We do not support `&[u8]` or `Reader` because the input must be an already allocated
@@ -180,11 +193,12 @@ impl<'d> Document<'d> {
     /// let doc = roxmltree::Document::parse("<e/>").unwrap();
     /// assert_eq!(doc.descendants().count(), 2); // root node + `e` element node
     /// ```
+    #[inline]
     pub fn parse(text: &str) -> Result<Document, Error> {
         parse(text)
     }
 
-    fn append(&mut self, parent_id: NodeId, kind: NodeKind<'d>, orig_pos: usize) -> NodeId {
+    fn append(&mut self, parent_id: NodeId, kind: NodeKind<'input>, range: Range) -> NodeId {
         let new_child_id = NodeId(self.nodes.len());
         self.nodes.push(NodeData {
             parent: Some(parent_id),
@@ -192,7 +206,7 @@ impl<'d> Document<'d> {
             next_sibling: None,
             children: None,
             kind,
-            orig_pos,
+            range,
         });
 
         let last_child_id = self.nodes[parent_id.0].children.map(|(_, id)| id);
@@ -212,39 +226,42 @@ impl<'d> Document<'d> {
 
         new_child_id
     }
-
-    fn get(&self, id: NodeId) -> Node {
-        Node { id, d: &self.nodes[id.0], doc: self }
-    }
 }
 
-struct Entity<'d>  {
-    name: &'d str,
-    value: StrSpan<'d>,
+struct Entity<'input>  {
+    name: &'input str,
+    value: StrSpan<'input>,
 }
 
-struct ParserData<'d> {
+struct ParserData<'input> {
     attrs_start_idx: usize,
     ns_start_idx: usize,
-    tmp_attrs: Vec<AttributeData<'d>>,
-    entities: Vec<Entity<'d>>,
+    tmp_attrs: Vec<AttributeData<'input>>,
+    entities: Vec<Entity<'input>>,
     buffer: TextBuffer,
     after_text: bool,
 }
 
 #[derive(Clone, Copy)]
-struct TagNameSpan<'d> {
-    prefix: StrSpan<'d>,
-    name: StrSpan<'d>,
+struct TagNameSpan<'input> {
+    prefix: StrSpan<'input>,
+    name: StrSpan<'input>,
+    span: StrSpan<'input>,
 }
 
-impl<'d> TagNameSpan<'d> {
+impl<'input> TagNameSpan<'input> {
+    #[inline]
     fn new_null() -> Self {
-        Self { prefix: StrSpan::from(""), name: StrSpan::from("") }
+        Self {
+            prefix: StrSpan::from(""),
+            name: StrSpan::from(""),
+            span: StrSpan::from(""),
+        }
     }
 
-    fn new(prefix: StrSpan<'d>, name: StrSpan<'d>) -> Self {
-        Self { prefix, name }
+    #[inline]
+    fn new(prefix: StrSpan<'input>, name: StrSpan<'input>, span: StrSpan<'input>) -> Self {
+        Self { prefix, name, span }
     }
 }
 
@@ -277,7 +294,7 @@ fn parse(text: &str) -> Result<Document, Error> {
         next_sibling: None,
         children: None,
         kind: NodeKind::Root,
-        orig_pos: 0,
+        range: 0..0,
     });
 
     doc.namespaces.push_ns(Some("xml"), NS_XML_URI.to_string());
@@ -298,65 +315,62 @@ fn parse(text: &str) -> Result<Document, Error> {
     Ok(doc)
 }
 
-fn process_tokens<'d>(
-    parser: xmlparser::Tokenizer<'d>,
+fn process_tokens<'input>(
+    parser: xmlparser::Tokenizer<'input>,
     entity_depth: u8,
     mut parent_id: NodeId,
-    tag_name: &mut TagNameSpan<'d>,
-    pd: &mut ParserData<'d>,
-    doc: &mut Document<'d>,
+    tag_name: &mut TagNameSpan<'input>,
+    pd: &mut ParserData<'input>,
+    doc: &mut Document<'input>,
 ) -> Result<(), Error> {
     for token in parser {
         let token = token?;
         match token {
-            xmlparser::Token::ProcessingInstruction(target, content) => {
-                let orig_pos = target.start() - 2; // jump before '<?'
+            xmlparser::Token::ProcessingInstruction { target, content, span } => {
                 let pi = NodeKind::PI(PI {
-                    target: target.to_str(),
-                    value: content.map(|v| v.to_str()),
+                    target: target.as_str(),
+                    value: content.map(|v| v.as_str()),
                 });
-                doc.append(parent_id, pi, orig_pos);
+                doc.append(parent_id, pi, span.range());
             }
-            xmlparser::Token::Comment(text) => {
-                let orig_pos = text.start() - 4; // jump before '<!--'
-                doc.append(parent_id, NodeKind::Comment(text.to_str()), orig_pos);
+            xmlparser::Token::Comment { text, span } => {
+                doc.append(parent_id, NodeKind::Comment(text.as_str()), span.range());
             }
-            xmlparser::Token::Text(text) |
-            xmlparser::Token::Whitespaces(text) => {
+            xmlparser::Token::Text { text } => {
                 process_text(text, parent_id, entity_depth, pd, doc)?;
             }
-            xmlparser::Token::Cdata(text) => {
-                let cow_str = Cow::Borrowed(text.to_str());
-                append_text(cow_str, parent_id, text.start(), pd.after_text, doc);
+            xmlparser::Token::Cdata { text, span } => {
+                let cow_str = Cow::Borrowed(text.as_str());
+                append_text(cow_str, parent_id, span.range(), pd.after_text, doc);
                 pd.after_text = true;
             }
-            xmlparser::Token::ElementStart(prefix, local) => {
-                if prefix.to_str() == "xmlns" {
+            xmlparser::Token::ElementStart { prefix, local, span } => {
+                if prefix.as_str() == "xmlns" {
                     let pos = err_pos_from_span(prefix);
                     return Err(Error::InvalidElementNamePrefix(pos));
                 }
 
-                *tag_name = TagNameSpan::new(prefix, local);
+                *tag_name = TagNameSpan::new(prefix, local, span);
             }
-            xmlparser::Token::Attribute((prefix, local), value) => {
-                process_attribute(entity_depth, prefix, local, value, pd, doc)?;
+            xmlparser::Token::Attribute { prefix, local, value, span } => {
+                process_attribute(entity_depth, prefix, local, value, span, pd, doc)?;
             }
-            xmlparser::Token::ElementEnd(end) => {
-                process_element(*tag_name, end, &mut parent_id, pd, doc)?;
+            xmlparser::Token::ElementEnd { end, span } => {
+                process_element(*tag_name, end, span, &mut parent_id, pd, doc)?;
             }
-            xmlparser::Token::EntityDeclaration(name, value) => {
-                if let xmlparser::EntityDefinition::EntityValue(value) = value {
-                    pd.entities.push(Entity { name: name.to_str(), value });
+            xmlparser::Token::EntityDeclaration { name, definition, .. } => {
+                if let xmlparser::EntityDefinition::EntityValue(value) = definition {
+                    pd.entities.push(Entity { name: name.as_str(), value });
                 }
             }
             _ => {}
         }
 
         match token {
-            xmlparser::Token::ProcessingInstruction(..) |
-            xmlparser::Token::Comment(..) |
-            xmlparser::Token::ElementStart(..) |
-            xmlparser::Token::ElementEnd(..) => {
+            xmlparser::Token::ProcessingInstruction { .. } |
+            xmlparser::Token::Comment { .. } |
+            xmlparser::Token::ElementStart { .. } |
+            xmlparser::Token::ElementEnd { .. } => {
                 pd.after_text = false;
             }
             _ => {}
@@ -366,20 +380,20 @@ fn process_tokens<'d>(
     Ok(())
 }
 
-fn process_attribute<'d>(
+fn process_attribute<'input>(
     entity_depth: u8,
-    prefix: StrSpan<'d>,
-    local: StrSpan<'d>,
-    value: StrSpan<'d>,
-    pd: &mut ParserData<'d>,
-    doc: &mut Document<'d>,
+    prefix: StrSpan<'input>,
+    local: StrSpan<'input>,
+    value: StrSpan<'input>,
+    token_span: StrSpan<'input>,
+    pd: &mut ParserData<'input>,
+    doc: &mut Document<'input>,
 ) -> Result<(), Error> {
-    let prefix_str = prefix.to_str();
-    let local_str = local.to_str();
-    let value_pos = value.start();
+    let range = token_span.range();
+    let value_range = value.range();
     let value = normalize_attribute(entity_depth, value, &pd.entities, &mut pd.buffer)?;
 
-    if prefix_str == "xmlns" {
+    if prefix.as_str() == "xmlns" {
         // The xmlns namespace MUST NOT be declared as the default namespace.
         if value == NS_XMLNS_URI {
             let pos = err_pos_from_qname(prefix, local);
@@ -391,7 +405,7 @@ fn process_attribute<'d>(
         // The prefix 'xml' is by definition bound to the namespace name
         // http://www.w3.org/XML/1998/namespace.
         // It MUST NOT be bound to any other namespace name.
-        if local_str == "xml" {
+        if local.as_str() == "xml" {
             if !is_xml_ns_uri {
                 let pos = err_pos_from_span(prefix);
                 return Err(Error::InvalidXmlPrefixUri(pos));
@@ -405,16 +419,16 @@ fn process_attribute<'d>(
         }
 
         // Check for duplicated namespaces.
-        if doc.namespaces.exists(pd.ns_start_idx, Some(local_str)) {
+        if doc.namespaces.exists(pd.ns_start_idx, Some(local.as_str())) {
             let pos = err_pos_from_qname(prefix, local);
-            return Err(Error::DuplicatedNamespace(local_str.to_string(), pos));
+            return Err(Error::DuplicatedNamespace(local.as_str().to_string(), pos));
         }
 
         // Xml namespace should not be added to the namespaces.
         if !is_xml_ns_uri {
-            doc.namespaces.push_ns(Some(local_str), value.into());
+            doc.namespaces.push_ns(Some(local.as_str()), value.into());
         }
-    } else if local_str == "xmlns" {
+    } else if local.as_str() == "xmlns" {
         // The xml namespace MUST NOT be declared as the default namespace.
         if value == NS_XML_URI {
             let pos = err_pos_from_span(local);
@@ -429,29 +443,29 @@ fn process_attribute<'d>(
 
         doc.namespaces.push_ns(None, value.into());
     } else {
-        let name = if prefix.is_empty() { local } else { prefix };
         pd.tmp_attrs.push(AttributeData {
-            name, prefix: prefix_str, local: local_str, value, value_pos
+            prefix, local, value, range, value_range
         });
     }
 
     Ok(())
 }
 
-fn process_element<'d>(
-    tag_name: TagNameSpan<'d>,
-    end_token: xmlparser::ElementEnd<'d>,
+fn process_element<'input>(
+    tag_name: TagNameSpan<'input>,
+    end_token: xmlparser::ElementEnd<'input>,
+    token_span: StrSpan<'input>,
     parent_id: &mut NodeId,
-    pd: &mut ParserData<'d>,
-    doc: &mut Document<'d>,
+    pd: &mut ParserData<'input>,
+    doc: &mut Document<'input>,
 ) -> Result<(), Error> {
     if tag_name.name.is_empty() {
         // May occur in XML like this:
         // <!DOCTYPE test [ <!ENTITY p '</p>'> ]>
         // <root>&p;</root>
 
-        if let xmlparser::ElementEnd::Close(prefix, local) = end_token {
-            return Err(Error::UnexpectedEntityCloseTag(err_pos_for_close_tag(prefix, local)));
+        if let xmlparser::ElementEnd::Close(..) = end_token {
+            return Err(Error::UnexpectedEntityCloseTag(err_pos_from_span(token_span)));
         } else {
             unreachable!("should be already checked by the xmlparser");
         }
@@ -460,38 +474,38 @@ fn process_element<'d>(
     let namespaces = resolve_namespaces(pd.ns_start_idx, *parent_id, doc);
     pd.ns_start_idx = doc.namespaces.len();
 
-    let attributes = resolve_attributes(pd.attrs_start_idx, namespaces.clone(), &mut pd.tmp_attrs, doc)?;
+    let attributes = resolve_attributes(pd.attrs_start_idx, namespaces.clone(),
+                                        &mut pd.tmp_attrs, doc)?;
     pd.attrs_start_idx = doc.attrs.len();
     pd.tmp_attrs.clear();
 
-    let tag_ns_uri = doc.namespaces.get_by_prefix(namespaces.clone(), tag_name.prefix.to_str());
     match end_token {
         xmlparser::ElementEnd::Empty => {
+            let tag_ns_uri = get_ns_by_prefix(doc, namespaces.clone(), tag_name.prefix)?;
             doc.append(*parent_id,
                 NodeKind::Element {
                     tag_name: ExpandedNameOwned {
                         ns: tag_ns_uri,
-                        name: tag_name.name.to_str(),
+                        prefix: tag_name.prefix.as_str(),
+                        name: tag_name.name.as_str(),
                     },
                     attributes,
                     namespaces,
                 },
-                orig_pos_from_tag_name(&tag_name)
+                tag_name.span.start()..token_span.end()
             );
         }
         xmlparser::ElementEnd::Close(prefix, local) => {
-            let prefix_str = prefix.to_str();
-            let local_str = local.to_str();
+            let prefix = prefix.as_str();
+            let local = local.as_str();
 
+            doc.nodes[parent_id.0].range.end = token_span.end();
             if let NodeKind::Element { ref tag_name, .. } = doc.nodes[parent_id.0].kind {
-                let parent_node = doc.get(*parent_id);
-                let parent_prefix = parent_node.resolve_tag_name_prefix().unwrap_or("");
-
-                if prefix_str != parent_prefix || local_str != tag_name.name {
+                if prefix != tag_name.prefix || local != tag_name.name {
                     return Err(Error::UnexpectedCloseTag {
-                        expected: gen_qname_string(parent_prefix, tag_name.name),
-                        actual: gen_qname_string(prefix_str, local_str),
-                        pos: err_pos_for_close_tag(prefix, local),
+                        expected: gen_qname_string(tag_name.prefix, tag_name.name),
+                        actual: gen_qname_string(prefix, local),
+                        pos: err_pos_from_span(token_span),
                     });
                 }
             }
@@ -503,16 +517,18 @@ fn process_element<'d>(
             }
         }
         xmlparser::ElementEnd::Open => {
+            let tag_ns_uri = get_ns_by_prefix(doc, namespaces.clone(), tag_name.prefix)?;
             *parent_id = doc.append(*parent_id,
                 NodeKind::Element {
                     tag_name: ExpandedNameOwned {
                         ns: tag_ns_uri,
-                        name: tag_name.name.to_str(),
+                        prefix: tag_name.prefix.as_str(),
+                        name: tag_name.name.as_str(),
                     },
                     attributes,
                     namespaces,
                 },
-                orig_pos_from_tag_name(&tag_name)
+                tag_name.span.start()..token_span.end()
             );
         }
     }
@@ -524,7 +540,7 @@ fn resolve_namespaces(
     start_idx: usize,
     parent_id: NodeId,
     doc: &mut Document,
-) -> Range<usize> {
+) -> Range {
     let mut tmp_parent_id = parent_id.0;
     while tmp_parent_id != 0 {
         let curr_id = tmp_parent_id;
@@ -553,18 +569,18 @@ fn resolve_namespaces(
     }
 }
 
-fn resolve_attributes<'d>(
+fn resolve_attributes<'input>(
     start_idx: usize,
-    namespaces: Range<usize>,
-    tmp_attrs: &mut [AttributeData<'d>],
-    doc: &mut Document<'d>,
-) -> Result<Range<usize>, Error> {
+    namespaces: Range,
+    tmp_attrs: &mut [AttributeData<'input>],
+    doc: &mut Document<'input>,
+) -> Result<Range, Error> {
     if tmp_attrs.is_empty() {
         return Ok(0..0);
     }
 
     for attr in tmp_attrs {
-        let ns = if attr.prefix == "xml" {
+        let ns = if attr.prefix.as_str() == "xml" {
             // The prefix 'xml' is by definition bound to the namespace name
             // http://www.w3.org/XML/1998/namespace.
             Some(doc.namespaces.xml_uri())
@@ -573,14 +589,16 @@ fn resolve_attributes<'d>(
             // always has no value.'
             None
         } else {
-            doc.namespaces.get_by_prefix(namespaces.clone(), attr.prefix)
+            get_ns_by_prefix(doc, namespaces.clone(), attr.prefix)?
         };
 
-        let attr_name = ExpandedNameOwned { ns, name: attr.local };
+        // We do not store attribute prefixes since `ExpandedNameOwned::prefix`
+        // is used only for closing tags matching during parsing.
+        let attr_name = ExpandedNameOwned { ns, prefix: "", name: attr.local.as_str() };
 
         // Check for duplicated attributes.
         if doc.attrs[start_idx..].iter().any(|attr| attr.name == attr_name) {
-            let pos = err_pos_from_span(attr.name);
+            let pos = err_pos_from_qname(attr.prefix, attr.local);
             return Err(Error::DuplicatedAttribute(attr.local.to_string(), pos));
         }
 
@@ -588,31 +606,31 @@ fn resolve_attributes<'d>(
             name: attr_name,
             // Takes a value from a slice without consuming the slice.
             value: mem::replace(&mut attr.value, Cow::Borrowed("")),
-            attr_pos: attr.name.start(),
-            value_pos: attr.value_pos,
+            range: attr.range.clone(),
+            value_range: attr.value_range.clone(),
         });
     }
 
     Ok(start_idx..doc.attrs.len())
 }
 
-fn process_text<'d>(
-    text: StrSpan<'d>,
+fn process_text<'input>(
+    text: StrSpan<'input>,
     parent_id: NodeId,
     entity_depth: u8,
-    pd: &mut ParserData<'d>,
-    doc: &mut Document<'d>,
+    pd: &mut ParserData<'input>,
+    doc: &mut Document<'input>,
 ) -> Result<(), Error> {
     // Add text as is if it has only valid characters.
-    if !text.to_str().bytes().any(|b| b == b'&' || b == b'\r') {
-        append_text(Cow::Borrowed(text.to_str()), parent_id, text.start(), pd.after_text, doc);
+    if !text.as_str().bytes().any(|b| b == b'&' || b == b'\r') {
+        append_text(Cow::Borrowed(text.as_str()), parent_id, text.range(), pd.after_text, doc);
         pd.after_text = true;
         return Ok(());
     }
 
-    fn _append_text(parent_id: NodeId, orig_pos: usize, pd: &mut ParserData, doc: &mut Document) {
+    fn _append_text(parent_id: NodeId, range: Range, pd: &mut ParserData, doc: &mut Document) {
         let cow_text = Cow::Owned(pd.buffer.to_str().to_owned());
-        append_text(cow_text, parent_id, orig_pos, pd.after_text, doc);
+        append_text(cow_text, parent_id, range, pd.after_text, doc);
         pd.after_text = true;
         pd.buffer.clear();
     }
@@ -653,7 +671,7 @@ fn process_text<'d>(
                 }
 
                 if !pd.buffer.is_empty() {
-                    _append_text(parent_id, text.start(), pd, doc);
+                    _append_text(parent_id, text.range(), pd, doc);
                 }
 
                 let mut parser = xmlparser::Tokenizer::from(fragment);
@@ -668,18 +686,18 @@ fn process_text<'d>(
     }
 
     if !pd.buffer.is_empty() {
-        _append_text(parent_id, text.start(), pd, doc);
+        _append_text(parent_id, text.range(), pd, doc);
     }
 
     Ok(())
 }
 
-fn append_text<'d>(
-    text: Cow<'d, str>,
+fn append_text<'input>(
+    text: Cow<'input, str>,
     parent_id: NodeId,
-    orig_pos: usize,
+    range: Range,
     after_text: bool,
-    doc: &mut Document<'d>,
+    doc: &mut Document<'input>,
 ) {
     if after_text {
         // Prepend to a previous text node.
@@ -696,7 +714,7 @@ fn append_text<'d>(
             }
         }
     } else {
-        doc.append(parent_id, NodeKind::Text(text), orig_pos);
+        doc.append(parent_id, NodeKind::Text(text), range);
     }
 }
 
@@ -718,17 +736,18 @@ fn parse_next_chunk<'a>(
 
     // Check for character/entity references.
     if c == b'&' {
+        let start = s.pos();
         match s.try_consume_reference() {
-            Some(Reference::CharRef(ch)) => {
+            Some(Reference::Char(ch)) => {
                 Ok(NextChunk::Char(ch))
             }
-            Some(Reference::EntityRef(name)) => {
+            Some(Reference::Entity(name)) => {
                 match entities.iter().find(|e| e.name == name) {
                     Some(entity) => {
                         Ok(NextChunk::Text(entity.value))
                     }
                     None => {
-                        let pos = s.gen_text_pos();
+                        let pos = s.gen_text_pos_from(start);
                         Err(Error::UnknownEntityReference(name.into(), pos))
                     }
                 }
@@ -745,21 +764,22 @@ fn parse_next_chunk<'a>(
 }
 
 // https://www.w3.org/TR/REC-xml/#AVNormalize
-fn normalize_attribute<'d>(
+fn normalize_attribute<'input>(
     entity_depth: u8,
-    text: StrSpan<'d>,
+    text: StrSpan<'input>,
     entities: &[Entity],
     buffer: &mut TextBuffer,
-) -> Result<Cow<'d, str>, Error> {
+) -> Result<Cow<'input, str>, Error> {
     if is_normalization_required(&text) {
         buffer.clear();
         _normalize_attribute(text, entities, entity_depth, buffer)?;
         Ok(Cow::Owned(buffer.to_str().to_owned()))
     } else {
-        Ok(Cow::Borrowed(text.to_str()))
+        Ok(Cow::Borrowed(text.as_str()))
     }
 }
 
+#[inline]
 fn is_normalization_required(text: &StrSpan) -> bool {
     // We assume that `&` indicates an entity or a character reference.
     // But in rare cases it can be just an another character.
@@ -774,7 +794,7 @@ fn is_normalization_required(text: &StrSpan) -> bool {
         }
     }
 
-    text.to_str().bytes().any(check)
+    text.as_str().bytes().any(check)
 }
 
 fn _normalize_attribute(
@@ -792,13 +812,14 @@ fn _normalize_attribute(
 
         if c != b'&' {
             s.advance(1);
-            buffer.push_from_attr(c, s.get_curr_byte());
+            buffer.push_from_attr(c, s.curr_byte().ok());
             continue;
         }
 
         // Check for character/entity references.
+        let start = s.pos();
         match s.try_consume_reference() {
-            Some(Reference::CharRef(ch)) => {
+            Some(Reference::Char(ch)) => {
                 for b in CharToBytes::new(ch) {
                     if entity_depth > 0 {
                         buffer.push_from_attr(b, None);
@@ -809,7 +830,7 @@ fn _normalize_attribute(
                     }
                 }
             }
-            Some(Reference::EntityRef(name)) => {
+            Some(Reference::Entity(name)) => {
                 if entity_depth > ENTITY_DEPTH {
                     let pos = s.gen_text_pos();
                     return Err(Error::EntityReferenceLoop(pos));
@@ -821,14 +842,14 @@ fn _normalize_attribute(
                         _normalize_attribute(entity.value, entities, entity_depth, buffer)?;
                     }
                     None => {
-                        let pos = s.gen_text_pos();
+                        let pos = s.gen_text_pos_from(start);
                         return Err(Error::UnknownEntityReference(name.into(), pos));
                     }
                 }
             }
             None => {
                 s.advance(1);
-                buffer.push_from_attr(c, s.get_curr_byte());
+                buffer.push_from_attr(c, s.curr_byte().ok());
             }
         }
     }
@@ -836,6 +857,45 @@ fn _normalize_attribute(
     Ok(())
 }
 
+fn get_ns_by_prefix(
+    doc: &Document,
+    range: Range,
+    prefix: StrSpan,
+) -> Result<Option<Uri>, Error> {
+    // Prefix CAN be empty when the default namespace was defined.
+    //
+    // Example:
+    // <e xmlns='http://www.w3.org'/>
+    let prefix_opt = if prefix.is_empty() { None } else { Some(prefix.as_str()) };
+
+    let uri = doc.namespaces[range].iter()
+        .find(|ns| ns.name == prefix_opt)
+        .map(|ns| ns.uri.clone());
+
+    match uri {
+        Some(v) => Ok(Some(v)),
+        None => {
+            if !prefix.is_empty() {
+                // If an URI was not found and prefix IS NOT empty than
+                // we have an unknown namespace.
+                //
+                // Example:
+                // <e random:a='b'/>
+                let pos = err_pos_from_span(prefix);
+                Err(Error::UnknownNamespace(prefix.as_str().to_string(), pos))
+            } else {
+                // If an URI was not found and prefix IS empty than
+                // an element or an attribute doesn't have a namespace.
+                //
+                // Example:
+                // <e a='b'/>
+                Ok(None)
+            }
+        }
+    }
+}
+
+#[inline]
 fn gen_qname_string(prefix: &str, local: &str) -> String {
     if prefix.is_empty() {
         local.to_string()
@@ -844,26 +904,17 @@ fn gen_qname_string(prefix: &str, local: &str) -> String {
     }
 }
 
+#[inline]
 fn err_pos_from_span(text: StrSpan) -> TextPos {
     Stream::from(text).gen_text_pos()
 }
 
+#[inline]
 fn err_pos_from_qname(prefix: StrSpan, local: StrSpan) -> TextPos {
     let err_span = if prefix.is_empty() { local } else { prefix };
     err_pos_from_span(err_span)
 }
 
-fn err_pos_for_close_tag(prefix: StrSpan, local: StrSpan) -> TextPos {
-    let mut pos = err_pos_from_qname(prefix, local);
-    pos.col -= 2; // jump before '</'
-    pos
-}
-
-fn orig_pos_from_tag_name(tag_name: &TagNameSpan) -> usize {
-    let span = if tag_name.prefix.is_empty() { tag_name.name } else { tag_name.prefix };
-    span.start() - 1 // jump before '<'
-}
-
 
 mod internals {
     /// Iterate over `char` by `u8`.
@@ -873,6 +924,7 @@ mod internals {
     }
 
     impl CharToBytes {
+        #[inline]
         pub fn new(c: char) -> Self {
             let mut buf = [0xFF; 4];
             c.encode_utf8(&mut buf);
@@ -887,6 +939,7 @@ mod internals {
     impl Iterator for CharToBytes {
         type Item = u8;
 
+        #[inline]
         fn next(&mut self) -> Option<Self::Item> {
             if self.idx < 4 {
                 let b = self.buf[self.idx as usize];
@@ -909,12 +962,14 @@ mod internals {
     }
 
     impl TextBuffer {
+        #[inline]
         pub fn new() -> Self {
             TextBuffer {
                 buf: Vec::with_capacity(32),
             }
         }
 
+        #[inline]
         pub fn push_raw(&mut self, c: u8) {
             self.buf.push(c);
         }
@@ -954,14 +1009,17 @@ mod internals {
             }
         }
 
+        #[inline]
         pub fn clear(&mut self) {
             self.buf.clear();
         }
 
+        #[inline]
         pub fn is_empty(&self) -> bool {
             self.buf.is_empty()
         }
 
+        #[inline]
         pub fn to_str(&self) -> &str {
             use std::str;
 
